# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_comment_data

__all__ = [
  "Comment"
]

class Comment(object):
  def __init__(self, data, comment_id=None, proxies={}):
    self.proxies = proxies
    if data is None and comment_id:
      self._data = data = get_comment_data(comment_id, proxies=proxies)
    else:
      self._data = data
    for field, body in self.data.items():
      if not hasattr(self, field):
        setattr(self, field, body)

  def __str__(self):
    if self.author:
      return '''Comment({}: "{}")'''.format(self.id,
                                            self.body.replace('\r','').strip().split('\n',1)[0][:27]+'...' \
                                              if len(self.body.replace('\r',''))>30 \
                                              else self.body.replace('\r','').strip().split('\n',1)[0]
                                           )
    else:
      return f'''Deleted comment({self.id} in image {self.image_id})'''

  @property
  def data(self):
    return self._data
       
  @property
  def url(self):
    return f"https://derpibooru.org/images/{self.image_id}#comment_{self.id}"

  def update(self):
    data = get_comment_data(self.id, proxies=self.proxies)

    if data:
      self._data = data




from derpibooru import sort

def test_sorting_methods():
  sorting_methods = {
    "created_at",
    "score",
    "relevance",
    "height",
    "comments",
    "random"
  }

  methods = sort.__dict__

  assert len(sorting_methods) == len(methods)
  assert sorting_methods == sort.methods

  for key, value in methods.items():
    assert key == value.upper()



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_image_data, get_image_faves, request as request_image
from .comments import Comments
from .tags import Tags
from .filters import system_filters

__all__ = [
  "Image"
]

class Image(object):
  """
  This class provides a thin wrapper around JSON data, mapping each value to
  its own property. Once instantiated the data is immutable so as to reflect
  the stateless nature of a REST API.
  For getting image by id field data should be None and image_id contains id.
  For getting current featured image field data should be None and image_id="featured"
  API key need for checking my:***
  """
  def __init__(self, data, image_id=None, key="", search_params={}, proxies={}):
    self.proxies = proxies
    # key needed for checking my:***
    if key:
      self.key = key if key else search_params['key']
    elif 'key' in search_params:
      self.key = search_params['key']
    else:
      self.key = ""
    self._params = search_params

    # Set image_id="featured" for get current featured image
    if data is None and image_id:
      self._data = data = get_image_data(image_id, proxies=proxies)
    else:
      self._data = data

    for field, body in data.items():
      if not hasattr(self, field):
        setattr(self, field, body) 

  def __str__(self):
    return f"Image({self.id})"

  @property
  def tags(self):
    return self.data["tags"]

  @property
  def representations(self):
    sizes = self.data["representations"].items()
    images = { image: url for image, url in sizes }

    return images

  @property
  def full(self):
    return self.representations["full"]

  @property
  def large(self):
    return self.representations["large"]

  @property
  def medium(self):
    return self.representations["medium"]

  @property
  def small(self):
    return self.representations["small"]

  @property
  def tall(self):
    return self.representations["tall"]

  @property
  def thumb(self):
    return self.representations["thumb"]

  @property
  def thumb_small(self):
    return self.representations["thumb_small"]

  @property
  def thumb_tiny(self):
    return self.representations["thumb_tiny"]

  @property
  def image(self):
    return self.data["view_url"]

  @property
  def faved_by(self):
    faved_by = "favourited_by_users"

    if not faved_by in self.data:
      if self.faves > 0:
        self._data[faved_by] = get_image_faves(self.id, proxies=self.proxies)
      else:
        self._data[faved_by] = []

    return self._data[faved_by]

  def comments(self):
    # filter_id used to get comments for any image
    return Comments(filter_id=system_filters["everything"], proxies=self.proxies).image_id(self.id)
       
  @property
  def url(self):
    return f"https://derpibooru.org/images/{self.id}"

  @property
  def data(self):
    return self._data

  def update(self):
    data = get_image_data(self.id, proxies=self.proxies)

    if data:
      self._data = data

  @property
  def artists(self):
    self_tag_list = '(' + ' || '.join(f"name:{tag}" for tag in self.tags) + ')'
    art_tag_list = '(category:origin, namespace:?* || name:artist needed || name:anonymous artist || name:kotobukiya)'
    tags = Tags(q=(self_tag_list,art_tag_list,), per_page=50, 
                limit=len(self_tag_list), proxies=self.proxies)
    for tag in tags:
      yield tag.name_in_namespace

  @property
  def rating(self):
    #all_ratings = {tag.name for tag in Tags(q={"category:rating"}, proxies=self.proxies)}
    all_ratings = {"safe","suggestive","questionable","explicit",
                   "semi-grimdark","grimdark","grotesque"}
    rating_tags = list(set(self.tags).intersection(all_ratings))
    return rating_tags

  @property
  def species(self):
    self_tag_list = '(' + ' || '.join(f"name:{tag}" for tag in self.tags) + ')'
    sp_tag_list = '(category:species || name:humanized || name:anthro centaur || name:bat alicorn)'
    tags = Tags(q=(self_tag_list,sp_tag_list,), per_page=50, 
                limit=len(self_tag_list), proxies=self.proxies)
    for tag in tags:
      yield tag.name_in_namespace

  @property
  def characters(self):
    self_tag_list = '(' + ' || '.join(f"name:{tag}" for tag in self.tags) + ')'
    ch_tag_list = '(category:character || category:oc)'
    tags = Tags(q=(self_tag_list,ch_tag_list,"-name:oc","-name:oc only"),
                per_page=50, limit=len(self_tag_list), proxies=self.proxies)
    for tag in tags:
      yield tag.name_in_namespace

  @property
  def spoiler(self):
    spoiler_tags = []
    self_tag_list = '(' + ' || '.join(f"name:{tag}" for tag in self.tags) + ')'
    sp_tag_list = '(category:spoiler || category:content-official)'
    tags = Tags(q=(self_tag_list,sp_tag_list,), per_page=50,
                limit=len(self_tag_list), proxies=self.proxies)
    aliases = []
    for tag in tags:
      if tag.category == "spoiler" and tag.name != "leak":
        spoiler_tags.append(tag.name_in_namespace)
      elif tag.category == "content-official" and tag.aliases:
        aliases.extend(tag.aliases)
      elif tag.category == "content-official":
        spoiler_tags.append(tag.name_in_namespace)
    if aliases:
      aliases_str = ' || '.join(f"slug:{tag}" for tag in set(aliases))
      tags = Tags(q=(aliases_str,), per_page=50, limit=len(aliases), proxies=self.proxies)
      for tag in tags:
        if tag.category == "spoiler" and tag.name != "leak":
          spoiler_tags.append(tag.name_in_namespace)
    for tag in sorted(set(spoiler_tags)):
      yield tag

  @property
  def source(self):
    if self.data['source_url']:
      return self.data['source_url']
    else:
      return self.url
  
  @property
  def upvoted(self):
    """
    Checking image in my:upvotes.
    """
    images = request_image({'key': self.key, 'filter_id': system_filters["everything"], 'per_page': 1,
                            'q': (f'id:{self.id}','my:upvotes')}, proxies=self.proxies)
    for img in images:
      return True
    return False
  
  @property
  def downvoted(self):
    """
    Checking image in my:downvotes.
    """
    images = request_image({'key': self.key, 'filter_id': system_filters["everything"], 'per_page': 1,
                            'q': (f'id:{self.id}','my:downvotes')}, proxies=self.proxies)
    for img in images:
      return True
    return False
  
  @property
  def uploaded(self):
    """
    Checking image in my:uploads.
    """
    images = request_image({'key': self.key, 'filter_id': system_filters["everything"], 'per_page': 1,
                            'q': (f'id:{self.id}','my:uploads')}, proxies=self.proxies)
    for img in images:
      return True
    return False
  
  @property
  def faved(self):
    """
    Checking image in my:faves.
    """
    images = request_image({'key': self.key, 'filter_id': system_filters["everything"], 'per_page': 1,
                            'q': (f'id:{self.id}','my:faves')}, proxies=self.proxies)
    for img in images:
      return True
    return False
  
  @property
  def watched(self):
    """
    Checking image in my:watches.
    """
    images = request_image({'key': self.key, 'filter_id': system_filters["everything"], 'per_page': 1,
                            'q': (f'id:{self.id}','my:faves')}, proxies=self.proxies)
    for img in images:
      return True
    return False
  
  def next(self):
    """
    Get next Image in Search().
    """
    parameters = {**self._params, 'per_page': 1, 'sf': 'created_at'}
    parameters['q'].add(f'id.lt:{self.id}')
    data = request_image(parameters, proxies=self.proxies)
    try:
      return Image(next(data),
                   search_params={**self._params,
                                  'key': self.key if self.key else self._params['key']},
                   proxies=self.proxies)
    except StopIteration:
      return self
  
  def prev(self):
    """
    Get previous Image in Search().
    """
    parameters = {**self._params, 'per_page': 1, 'sf': 'created_at'}
    parameters['q'].add(f'id.gt:{self.id}')
    parameters['sd'] = 'desc' if self._params['sd']=='asc' else 'asc'
    data = request_image(parameters, proxies=self.proxies)
    try:
      return Image(next(data),
                   search_params={**self._params,
                                  'key': self.key if self.key else self._params['key']},
                   proxies=self.proxies)
    except StopIteration:
      return self



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_forums, get_forum_data, \
                     get_topics, url_topics, get_topic_data, \
                     get_posts, url_posts
from .helpers import join_params, set_limit, destructive_slug
from .post import Post

__all__ = [
  "Forums",
  "Forum",
  "Topics",
  "Topic",
  "Posts"
]

class Forums(object):
  def __init__(self, limit=50, per_page=25, page=1, proxies={}):
    self.proxies = proxies
    self._params = {
      "per_page": set_limit(per_page),
      "page": set_limit(page)
    }      
    self._limit = set_limit(limit)
    self._search = get_forums(self._params, self._limit, proxies=self.proxies)
  
  def __iter__(self):
    """
    Make Forums() iterable so that new search results can be lazily generated
    for performance reasons.
    """
    return self

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of Forums().
    """
    return self._params

  @property
  def url(self):
    """
    Returns a standart URL of avaliable forums list
    """
    return "https://derpibooru.org/forums"

  def limit(self, limit):
    """
    Set absolute limit on number of forums to return, or set to None to return
    as many results as needed; default 50 forums. This limit on app-level.
    """
    params = join_params(self.parameters, {"limit": limit, "proxies": self.proxies})

    return self.__class__(**params)

  def get_page(self,page):
    """
    Set page for gets result of search.
    """
    params = join_params(self.parameters, {"page": set_limit(page),
                                           "limit": self._limit,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def per_page(self,limit):
    """
    Set absolute limit on number of forums to get, or set to None to return
    defaulting 25 forums; max 50 forums. This limit on API-level.
    """
    params = join_params(self.parameters, {"per_page": set_limit(limit),
                                           "limit": self._limit,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def __next__(self):
    """
    Returns a result wrapped in a new instance of Forum().
    """
    return Forum(next(self._search), proxies=self.proxies)

class Forum(object):
  """
  This class provides a thin wrapper around JSON data, mapping each value to
  its own property. Once instantiated the data is immutable so as to reflect
  the stateless nature of a REST API.
  """
  def __init__(self, data, short_name=None, proxies={}):
    self.proxies = proxies

    if data is None and short_name:
      self._data = data = get_forum_data(short_name, proxies=proxies)
    else:
      self._data = data

    for field, body in data.items():
      if not hasattr(self, field):
        setattr(self, field, body) 

  def __str__(self):
    return f"Forum({self.name})"
       
  @property
  def url(self):
    return f"https://derpibooru.org/forums/{self.short_name}"

  @property
  def data(self):
    return self._data

  def update(self):
    data = get_forum_data(self.short_name, proxies=self.proxies)

    if data:
      self._data = data

  def topics(self, limit=50, per_page=25, page=1):
    return Topics(self.short_name, limit=limit,
                  per_page=per_page, page=page, proxies=self.proxies)

class Topics(object):
  def __init__(self, forum_short_name, limit=50, per_page=25, page=1, proxies={}):
    self.proxies = proxies
    self.forum_short_name = forum_short_name
    self._params = {
      "per_page": set_limit(per_page),
      "page": set_limit(page)
    }      
    self._limit = set_limit(limit)
    self._search = get_topics(forum_short_name, self._params,
                              self._limit, proxies=self.proxies)
  
  def __iter__(self):
    """
    Make Topics() iterable so that new search results can be lazily generated
    for performance reasons.
    """
    return self

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of Topics().
    """
    return self._params

  @property
  def url(self):
    """
    Returns a search URL built on set parameters. Example based on default
    parameters:

    https://derpibooru.org/forums/forum?page=1&per_page=25
    """
    return url_topics(self.forum_short_name, self.parameters)

  def limit(self, limit):
    """
    Set absolute limit on number of topics to return, or set to None to return
    as many results as needed; default 50 topics. This limit on app-level.
    """
    params = join_params(self.parameters, {"limit": limit, "proxies": self.proxies})

    return self.__class__(**params)

  def get_page(self,page):
    """
    Set page for gets result of search.
    """
    params = join_params(self.parameters, {"page": set_limit(page),
                                           "limit": self._limit,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def per_page(self,limit):
    """
    Set absolute limit on number of topics to get, or set to None to return
    defaulting 25 topics; max 50 topics. This limit on API-level.
    """
    params = join_params(self.parameters, {"per_page": set_limit(limit),
                                           "limit": self._limit,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def __next__(self):
    """
    Returns a result wrapped in a new instance of Topic().
    """
    return Topic(next(self._search), forum_short_name=self.forum_short_name,
                 proxies=self.proxies)

class Topic(object):
  """
  This class provides a thin wrapper around JSON data, mapping each value to
  its own property. Once instantiated the data is immutable so as to reflect
  the stateless nature of a REST API.
  """
  def __init__(self, data, forum_short_name=None, topic_name=None, slug=True, proxies={}):
    self.proxies = proxies
    self.forum_short_name = forum_short_name

    if data is None and forum_short_name and topic_name:
      if not slug:
        topic_slug = destructive_slug(topic_name)
      else:
        topic_slug = topic_name
      self._data = data = get_topic_data(forum_short_name, topic_slug, proxies=proxies)
    else:
      self._data = data

    for field, body in data.items():
      if not hasattr(self, field):
        setattr(self, field, body) 

  def __str__(self):
    return f"Topic({self.title})"
       
  @property
  def url(self):
    if self.forum_short_name:
      return f"https://derpibooru.org/forums/{self.forum_short_name}/topics/{self.slug}"

  @property
  def data(self):
    return self._data

  def update(self):
    if self.forum_short_name:
      data = get_topic_data(self.forum_short_name, self.slug, proxies=self.proxies)
      if data:
        self._data = data

  def posts(self, forum_short_name="", limit=50, per_page=25, page=1):
    if not forum_short_name:
      forum_short_name = self.forum_short_name
    return Posts(forum_short_name, self.slug, slug=False, limit=limit,
                 per_page=per_page, page=page, proxies=self.proxies)

class Posts(object):
  def __init__(self, forum_short_name, topic_name, slug=True, limit=50,
               per_page=25, page=1, proxies={}):
    self.proxies = proxies
    self.forum_short_name = forum_short_name
    if slug:
      self.topic_slug = destructive_slug(topic_name)
    else:
      self.topic_slug = topic_name
    self._params = {
      "per_page": set_limit(per_page),
      "page": set_limit(page)
    }
    self._limit = set_limit(limit)
    self._search = get_posts(self._params, forum_short_name=self.forum_short_name,
                             topic_slug=self.topic_slug, limit=self._limit,
                             proxies=self.proxies)
  
  def __iter__(self):
    """
    Make Posts() iterable so that new search results can be lazily generated
    for performance reasons.
    """
    return self

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of Posts().
    """
    return self._params

  @property
  def url(self):
    """
    Returns a search URL built on set parameters. Example based on default
    parameters:

    https://derpibooru.org/forums/forum/topics/topic?page=1&per_page=25
    """
    return url_posts(self.forum_short_name, self.topic_slug, self.parameters)

  def limit(self, limit):
    """
    Set absolute limit on number of posts to return, or set to None to return
    as many results as needed; default 50 posts. This limit on app-level.
    """
    params = join_params(self.parameters, {"limit": limit, "proxies": self.proxies})

    return self.__class__(**params)

  def get_page(self,page):
    """
    Set page for gets result of search.
    """
    params = join_params(self.parameters, {"page": set_limit(page),
                                           "limit": self._limit,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def per_page(self,limit):
    """
    Set absolute limit on number of posts to get, or set to None to return
    defaulting 25 posts; max 50 posts. This limit on API-level.
    """
    params = join_params(self.parameters, {"per_page": set_limit(limit),
                                           "limit": self._limit,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def __next__(self):
    """
    Returns a result wrapped in a new instance of Post().
    """
    return Post(next(self._search), proxies=self.proxies)


# -*- coding: utf-8 -*-

"""
Derpibooru API bindings
~~~~~~~~~~~~~~~~~~~~~~~

Python bindings for Derpibooru's API

Typical usage:

>>> from derpibooru import Search, sort
>>> for image in Search().sort_by(sort.SCORE):
...   print(image.url)

Full API Documentation is found at <https://derpibooru.org/pages/api>.

Library documentation is found at <https://github.com/Atronar/DerPyBooruPh>.

"""

__title__ = "DerPyBooru"
__version__ = "0.10.0"
__author__ = "Joshua Stone"
__license__ = "Simplified BSD Licence"
__copyright__ = "Copyright (c) 2014, Joshua Stone; 2020, ATroN"

from .search import Search, Related
from .image import Image
from .post_image import PostImage
from .comments import Comments
from .comment import Comment
from .tags import Tags
from .tag import Tag
from .profile import Profile
from .filters import Filters, Filter
from .galleries import Galleries
from .gallery import Gallery
from .posts import SearchPosts
from .post import Post
from .forums import Forums, Forum, Topics, Topic, Posts
from .query import query
from .sort import sort
from .user import user

__all__ = [
  "Search", "Related",
  "Image",
  "PostImage",
  "Comments",
  "Comment",
  "Tags",
  "Tag",
  "Profile",
  "Filters", "Filter",
  "Galleries",
  "Gallery",
  "SearchPosts",
  "Post",
  "Forums", "Forum", "Topics", "Topic", "Posts",
  "query",
  "sort",
  "user"
]



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__all__ = [
  "sort"
]

class Sort(object):
  def __init__(self):
    for method in self.methods:
      setattr(self, method.upper(), method)

  @property
  def methods(self):
    sorting_methods = {
      "created_at",
      "updated_at",
      "first_seen_at",
      "aspect_ratio",
      "faves",
      "upvotes",
      "downvotes",
      "score",
      "wilson_score",
      "relevance",
      "width",
      "height",
      "comments",
      "tag_count",
      "random"
    }

    return sorting_methods

sort = Sort()




# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from urllib.parse import quote_plus

__all__ = [
  "tags",
  "search_comments_fields",
  "api_key",
  "validate_filter",
  "validate_url",
  "validate_description",
  "sort_format",
  "user_option",
  "format_params",
  "join_params",
  "set_limit",
  "set_distance",
  "slugging_tag",
  "destructive_slug"
]

from .sort import sort
from .user import user

def tags(q):
  if isinstance(q, str):
    q = q.split(',')
  tags = {str(tag).strip() for tag in q if tag}

  return tags if tags else set()

def search_comments_fields(q, author="", body="", created_at="", comment_id="", \
                           image_id="", my=None, user_id=""):
  if ',' in q:
    q = q.split(',')

  if isinstance(body,str):
     if body:
       tags = body.split(',');
     else:
       tags = []
  elif hasattr(body,"__iter__"): # iterable: lists, sets, etc.
    tags = list(body)
  else:
    tags = []

  for tag in q:
    if tag:
      tag = str(tag).strip()
      if tag.startswith("author:"):
        if not author:
          author = tag.replace("author:","",1)
      elif tag.startswith("created_at:"):
        if not created_at:
          created_at = tag.replace("created_at:","",1)
      elif tag.startswith("comment_id:"):
        if not comment_id:
          comment_id = tag.replace("comment_id:","",1)
      elif tag.startswith("image_id:"):
        if not image_id:
          image_id = tag.replace("image_id:","",1)
      elif tag.startswith("user_id:"):
        if not user_id:
          user_id = tag.replace("user_id:","",1)
      elif tag.startswith("-my:"): # -my:comments
        if my is None: 
          if tag == "-my:comments":
            my = False
      elif tag.startswith("my:"): # my:comments
        if not my: 
          if tag == "my:comments":
            my = True
      else:
        tags.append(tag)

  if author:
    tags.append(f"author:{author}")
  if created_at:
    tags.append(f"created_at:{created_at}")
  if comment_id:
    tags.append(f"id:{comment_id}")
  if image_id:
    tags.append(f"image_id:{image_id}")
  if user_id:
    tags.append(f"user_id:{user_id}")
  if my is not None:
    if my:
      tags.append("my:comments")
    else:
      tags.append("-my:comments")

  tags = set(tags)

  return tags if tags else set()

def api_key(api_key):
  return str(api_key) if api_key else ""

def validate_filter(filter_id):
  # is it always an number?
  return str(filter_id) if filter_id else None

def validate_url(url):
  if url:
    norm_url = str(url).strip().strip('/')
    if any(norm_url.startswith(scheme) for scheme in {"http://", "https://"}):
      return norm_url
    elif "://" in norm_url:
      return ""
    else:
      return f"http://{norm_url}"
  else:
    return ""

def validate_description(string):
  if len(string.encode('utf-8'))<50000:
    return string
  else:
    n=50000
    while True:
      try:
        cutted_string = string.encode('utf-8')[:n].decode('utf-8')
        return cutted_string
      except UnicodeDecodeError:
        n -= 1

def sort_format(sf):
  if sf not in sort.methods:
    raise AttributeError(sf)
  else:
    return sf

def user_option(option):
  if option: 
    if option not in user.options:
      raise AttributeError(option)
    else:
      return option
  else:
    return ""

def format_params(params):
  p = {}

  for key, value in params.items():
    if key == "q":
      p["q"] = ",".join(value) if value else "*"
    elif key == "reverse_url" and value:
      p["url"] = value
    elif key == "distance":
      if "reverse_url" in params and params["reverse_url"]:
        p[key] = value
    elif value:
      p[key] = value

  return p

def format_params_url_galleries(params):
  p = {}

  for key, value in params.items():
    if key == "q":
      q = ",".join(value) if value else "*"
      q = (i.strip() for i in q.split(","))
      for tag in q:
        if tag.startwith("title:"):
          p["gallery[title]"] = tag[6:]
        elif tag.startwith("description:"):
          p["gallery[description]"] = tag[12:]
        elif tag.startwith("user:"):
          p["gallery[creator]"] = tag[5:]
        elif tag.startwith("image_ids:"):
          p["gallery[include_image]"] = tag[10:]
    elif value:
      p[key] = value

  return p

def join_params(old_params, new_params):
  new_dict = {**old_params, **new_params}

  return new_dict

def set_limit(limit):

  if limit is not None:
    l = int(limit)
  else:
    l = None

  return l

def set_distance(distance):
  if distance:
    try:
      d = float(distance)
      if d < 0.2:
        d = 0.2
      elif d >= 1:
        d = 0.5
    except:
      d = 0.25
  else:
    d = 0.25
  return d

def slugging_tag(tag):
  slug = tag.strip().lower()
  do_slug = False
  for char in slug:
    if char in '/\\:.+ ':
      do_slug = True
      break
  if not do_slug and '-' in slug:
    for char in ('-dash-','-fwslash','-bwslash-','-colon-','-dot-','-plus-','stop'):
      if char in slug and char != 'stop':
        break
      elif char == 'stop':
        do_slug = True
  if do_slug:
    for i,j in {'-':'-dash-', '/':'-fwslash', '\\':'-bwslash-',
                ':':'-colon-', '.':'-dot-', '+':'-plus-'}.items():
      slug = slug.replace(i,j)
    slug = quote_plus(slug)
  return slug

def destructive_slug(string):
  output = string
  for char in string:
    if ord(char) not in range(ord(" "),ord("~")) or char=="'":
      output = output.replace(char,"")
    elif ord(char) not in range(ord("a"),ord("z")) \
     and ord(char) not in range(ord("A"),ord("Z")) \
     and ord(char) not in range(ord("0"),ord("9")):
      output = output.replace(char,"-")
  while "--" in output:
    output = output.replace("--","-")
  output = output.strip("-").lower()
  return output


# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__all__ = [
  "query"
]

class Query_Field(object):
  def __init__(self, name, is_neg=False):
    self.name = name
    self.is_neg = is_neg

  def __neg__(self):
    return self.__class__(self.name, is_neg=True)

class Equal(Query_Field):
  def __eq__(self, value):
    if value:
      return f"{'-' if self.is_neg else ''}{self.name}:{value}"
    else:
      raise ValueError(value)

  def __gt__(self, value):
    raise AttributeError("gt")

  def __lt__(self, value):
    raise AttributeError("lt")

  def __ge__(self, value):
    raise AttributeError("ge")

  def __le__(self, value):
    raise AttributeError("le")

 
class Comparable(Query_Field):
  def op(self, op, value):
    try:
      float(value)
      return f"{'-' if self.is_neg else ''}{self.name}.{op}:{value}"
    except:
      raise ValueError(value)
 
  def __eq__(self, value):
    return self.op("eq", value)
 
  def __gt__(self, value):
    return self.op("gt", value)
 
  def __lt__(self, value):
    return self.op("lt", value)
 
  def __ge__(self, value):
    return self.op("gte", value)
 
  def __le__(self, value):
    return self.op("lte", value) 


class Query(object):
  def __init__(self):
    for field in self.equal:
      setattr(self, field, Equal(field))

    for field in self.comparable:
      setattr(self, field, Comparable(field))

  def __neg__(self):
    return self.__class__()

  # literal, ngram
  @property
  def equal(self):
    return {"alias_of", "aliased", "aliases", "analyzed_name", "author", "body",
            "category", "description", "faved_by", "forum_id", "gallery_id", "image_ids",
            "implied_by", "implies", "my", "name", "name_in_namespace", "namespace",
            "original_format", "short_description", "slug", "source_url", "subject",
            "orig_sha512_hash", "sha512_hash", "topic_id", "uploader", "watcher_ids"}

  # int, float, date
  @property
  def comparable(self):
    return {"aspect_ratio", "comment_count", "created_at", "downvotes", "faved_by_id",
            "faves", "first_seen_at", "height", "id", "image_count", "image_id",
            "images", "score", "tag_count", "updated_at", "uploader_id", "upvotes",
            "user_id", "watcher_count", "width", "wilson_score"}

  @property
  def images_attr(self):
    return ("aspect_ratio", "comment_count", "created_at", "description", "downvotes",
            "faved_by", "faved_by_id", "faves", "first_seen_at", "gallery_id", "height",
            "id", "my", "source_url", "orig_sha512_hash", "original_format", "score",
            "sha512_hash", "tag_count", "updated_at", "uploader", "uploader_id",
            "upvotes", "width", "wilson_score")

  @property
  def comments_attr(self):
    return ("author", "body", "created_at", "id", "image_id", "my", "user_id")

  @property
  def tags_attr(self):
    return ("alias_of", "aliased", "aliases", "analyzed_name", "category",
            "description", "id", "images", "implied_by", "implies", "name",
            "name_in_namespace", "namespace", "short_description", "slug")

  @property
  def galleries_attr(self):
    return ("created_at", "description", "id", "image_count", "image_ids", "title",
            "updated_at", "user", "watcher_count", "watcher_ids")

  @property
  def posts_attr(self):
    return ("author", "body", "created_at", "forum_id", "id", "my",
            "subject", "topic_id", "updated_at", "user_id")

query = Query()
